#ifndef GRAPHENGINE_NAMESPACE_H_
#define GRAPHENGINE_NAMESPACE_H_

#ifndef GRAPHENGINE_IMPL_NAMESPACE
  /**
   * Namespace of public, non-inline symbols.
   * Define when embedding graphengine into another library.
   */
  #define GRAPHENGINE_IMPL_NAMESPACE impl
#endif

#endif // GRAPHENGINE_NAMESPACE_H_
